﻿using System;
using System.IO;
using System.Collections.Generic;
using System.Collections;

namespace MCNGen {
    
    /// <summary>
    /// Generates random names using Markov chains. 
    /// Based on http://www.siliconcommandergames.com/MarkovNameGenerator.htm
    /// 
    /// Author: Joonas Rajala
    /// </summary>
    public class NameGenerator {

        /// <summary>
        /// Generates random names based on the given list of strings
        /// and returns the result.
        /// </summary>
        /// <param name="samples">List of sample names.</param>
        /// <param name="order">Higher order means that the generated words are more like the samples but it limits the maximum amount of possible combinations</param>
        /// <param name="minLength">Minimum length for the generated names.</param>
        /// <param name="amount">Desired amount of names to generate. Might fall short if not enough samples.</param>
        /// <returns></returns>
        public List<string> Run(List<string> samples, int order, int minLength, int amount) {

            List<string> generated = new List<string>();
            Markov markovNameGen = new Markov(samples, order, minLength);
            string name;

            while (generated.Count <= amount) {
                if (markovNameGen.NextName() == null) {
                    //Console.Write("Generated " + generated.Count + " names, ");
                    break;
                }
                name = markovNameGen.NextName();

                if (name != null && !samples.Contains(name)) {
                    generated.Add(name);
                }
            }
            generated.Sort();

            return generated;

        }

        /// <summary>
        /// Generates random names based on the given list of strings
        /// and returns the result. 
        /// </summary>
        /// <param name="samples">Array of sample names</param>
        /// <param name="order">see http://en.wikipedia.org/wiki/Markov_chain </param>
        /// <param name="minLength">Minimum length for the generated names.</param>
        /// <param name="amount">Desired amount of names to generate. Might fall short if not enough samples.</param>
        /// <returns></returns>
        public List<string> Run(string[] samples, int order, int minLength, int amount) {
            List<string> s = new List<string>();
            for (int i = 0; i < samples.Length; i++) {
                s.Add(samples[i]);
            }

            return Run(s, order, minLength, amount);
        }

        /// <summary>
        /// Generates random names based on the given list of strings
        /// and returns the result.
        /// </summary>
        /// <param name="samples">Comma or new line seperated sample names</param>
        /// <param name="order">see http://en.wikipedia.org/wiki/Markov_chain </param>
        /// <param name="minLength">Minimum length for the generated names.</param>
        /// <param name="amount">Desired amount of names to generate. Might fall short if not enough samples.</param>
        /// <returns></returns>
        public List<string> Run(string samples, int order, int minLength, int amount) {

            string[] s = samples.Split(new string[] { ",", "\n", "\r\n" }, StringSplitOptions.RemoveEmptyEntries);

            List<string> asd = new List<string>();
            for (int i = 0; i < s.Length; i++) {
                asd.Add(s[i]);
            }

            return asd;
        }        

        /// <summary>
        /// Inner class for the name generator.
        /// 
        /// Generates names based on a list of sample strings using Markov Chains.
        /// </summary>
        private class Markov {

            private Dictionary<string, List<char>> chains = new Dictionary<string, List<char>>();
            private List<string> samples = new List<string>();
            private ArrayList used = new ArrayList();
            private Random rnd = new Random();
            private int order;
            private int minLength;
            private int retries = 0;
            private int maxRetries = 500000;

            /// <summary>
            /// Class constructor. Creates the token dictionary from the sample names.
            /// </summary>
            /// <param name="sampleNames">List of sample strings</param>
            /// <param name="order">see http://en.wikipedia.org/wiki/Markov_chain </param>
            /// <param name="minLength">Minmum length of the generated words</param>
            public Markov(List<string> sampleNames, int order, int minLength) {

                if (order < 1) this.order = 1;
                if (minLength < 1) this.minLength = 1;

                this.order = order;
                this.minLength = minLength;

                foreach (string s in sampleNames) {
                    string upper = s.Trim().ToUpper();

                    if (upper.Length > order) {
                        samples.Add(upper);
                    }
                }

                foreach (string word in samples) {
                    for (int letter = 0; letter < word.Length - order; letter++) {
                        string token = word.Substring(letter, order);
                        List<char> entry = null;
                        if (chains.ContainsKey(token))
                            entry = chains[token];
                        else {
                            entry = new List<char>();
                            chains[token] = entry;
                        }
                        entry.Add(word[letter + order]);
                    }
                }
            }

            /// <summary>
            /// Generate a random name
            /// </summary>
            /// <returns>random name</returns>
            public string NextName() {
                string s = "";
                do {
                    retries += 1;
                    if (retries >= maxRetries) {
                        return null;
                    }

                    int n = rnd.Next(samples.Count);
                    int nameLength = samples[n].Length;
                    s = samples[n].Substring(rnd.Next(0, samples[n].Length - order), order);
                    while (s.Length < nameLength) {
                        string token = s.Substring(s.Length - order, order);
                        char c = GetLetter(token);
                        if (c != '?') {
                            s += GetLetter(token);
                        } else break;
                    }

                    if (s.Contains(" ")) {
                        string[] tokens = s.Split(' ');
                        s = "";

                        for (int t = 0; t < tokens.Length; t++) {
                            if (tokens[t] == "") continue;

                            if (tokens[t].Length == 1) tokens[t] = tokens[t].ToUpper();

                            else tokens[t] = tokens[t].Substring(0, 1) + tokens[t].Substring(1).ToLower();

                            if (s != "") s += " ";
                            s += tokens[t];
                        }
                    } else s = s.Substring(0, 1) + s.Substring(1).ToLower();
                }
                while (used.Contains(s) || s.Length < minLength);
                used.Add(s);
                return s;

            }

            /// <summary>
            /// Reset the used names
            /// </summary>
            public void Reset() {
                used.Clear();
            }

            /// <summary>
            /// Get a random letter from the chain
            /// </summary>
            /// <param name="token"></param>
            /// <returns>A random letter</returns>
            private char GetLetter(string token) {
                if (!chains.ContainsKey(token))
                    return '?';
                List<char> letters = chains[token];
                int n = rnd.Next(letters.Count);
                return letters[n];
            }
        }
    }
}




